/** @file imagenii.c
    @brief Process NIfTI and Analyze images with IMG structure.
    @copyright (c) Turku PET Centre
    @author Vesa Oikonen
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include "tpcimage.h"
/*****************************************************************************/

/*****************************************************************************/
/** Read NIfTI-1 into IMG data structure.
    @remark Only few header fields are read.
    @sa imgInit, imgRead, imgWrite, imgFree
    @return enum tpcerror (TPCERROR_OK when successful).
 */
int imgReadNifti1(
  /** Pointer to image structure. Any previous contents removed. */
  IMG *img,
  /** Pointer to the database name. Path not accepted. */
  const char *fname,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {printf("%s(..., %s, ...)\n", __func__, fname); fflush(stdout);}

  if(img==NULL || strnlen(fname, 2)<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return(TPCERROR_FAIL);
  }

  /* Delete any old contents in data structure */
  imgFree(img);


  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write NIfTI-1 from IMG data structure.

   IMG field oformat (or secondarily format) determines whether NIfTI is written in single
   file format (*.nii) or dual file format (*.hdr and *.img).

   SIF file is saved to store frame times, if frame times are available.

   @sa imgInit, imgRead, imgWrite, imgFree
   @return enum tpcerror (TPCERROR_OK when successful).
 */
int imgWriteNifti1(
  /** Pointer to image structure. Any previous contents removed. */
  IMG *img,
  /** Pointer to the database name. Path not accepted. */
  const char *fname,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {printf("%s(..., %s, ...)\n", __func__, fname); fflush(stdout);}

  if(img==NULL || strnlen(fname, 2)<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return(TPCERROR_FAIL);
  }
  if(!imgHasData(img)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return(TPCERROR_NO_DATA);
  }

  /* Get database name, in case file name was given with extension */
  if(verbose>1) {printf("get dbname\n"); fflush(stdout);}
  char basename[FILENAME_MAX];
  strncpyCleanSpaces(basename, fname, FILENAME_MAX); 
  char *cptr=filenameGetExtension(basename);
  if(cptr!=NULL) {
    if(strcasecmp(cptr, ".hdr")==0) *cptr=(char)0;
    else if(strcasecmp(cptr, ".nii")==0) *cptr=(char)0;
    else if(strcasecmp(cptr, ".img")==0) *cptr=(char)0;
    else if(strcasecmp(cptr, ".sif")==0) *cptr=(char)0;
    else if(strcasecmp(cptr, ".")==0) *cptr=(char)0;
  }
  /* Check that there is something left after path */
  char basenamewopath[FILENAME_MAX];
  strcpy(basenamewopath, basename); filenameRmPath(basenamewopath);
  if(strlen(basenamewopath)<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_FILENAME);
    return(TPCERROR_INVALID_FILENAME);
  }
  /* If we have path, then create it */
  char filepath[FILENAME_MAX];
  strcpy(filepath, basename); filenameRmFile(filepath);
  if(pathCreate(filepath)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_FILENAME);
    return(TPCERROR_INVALID_FILENAME);
  }

  /* Determine the output format */
  if(verbose>1) {printf("determine output format\n"); fflush(stdout);}
  if(img->oformat==IMG_FORMAT_UNKNOWN) {
    char *cptr=filenameGetExtension(fname);
    if(cptr!=NULL && strcasecmp(cptr, ".nii")==0) img->oformat=IMG_FORMAT_NIFTI_1S;
    else img->oformat=img->format;
  }
  if(img->oformat!=IMG_FORMAT_NIFTI_1D && img->oformat!=IMG_FORMAT_NIFTI_1S) 
    img->oformat=IMG_FORMAT_NIFTI_1S;

  /* Create output file names */
  if(verbose>1) {printf("determine output file names\n"); fflush(stdout);}
  char hdrfile[FILENAME_MAX], imgfile[FILENAME_MAX], siffile[FILENAME_MAX];
  hdrfile[0]=imgfile[0]=siffile[0]=(char)0;
  if(img->oformat==IMG_FORMAT_NIFTI_1D) {
    snprintf(hdrfile, FILENAME_MAX, "%s.hdr", basename);
    snprintf(imgfile, FILENAME_MAX, "%s.img", basename);
  } else if(img->oformat==IMG_FORMAT_NIFTI_1S) {
    snprintf(imgfile, FILENAME_MAX, "%s.nii", basename);
  }
  if(imgHasTimes(img)) snprintf(siffile, FILENAME_MAX, "%s.sif", basename);

  /* Fill the header structure */
  NIFTI_DSR dsr;
  int ret=imgSetNifti1Header(img, basenamewopath, &dsr, verbose-1);
  if(ret!=TPCERROR_OK) {
    statusSet(status, __func__, __FILE__, __LINE__, ret);
    return(ret);
  }

  /* Delete previous NIfTI */
  if(verbose>1) {printf("removing any previous files\n"); fflush(stdout);}
  if(fileExist(hdrfile)) remove(hdrfile);
  if(fileExist(imgfile)) remove(imgfile);
  //if(fileExist(siffile)) remove(siffile); // NO, not this!

  /* Write the header */
  if(verbose>1) {printf("writing the header\n"); fflush(stdout);}
  if(img->oformat==IMG_FORMAT_NIFTI_1D) {
    ret=niftiWriteHeader(hdrfile, &dsr, verbose-1);
  } else if(img->oformat==IMG_FORMAT_NIFTI_1S) {
    ret=niftiWriteHeader(imgfile, &dsr, verbose-1);
  }
  if(ret!=TPCERROR_OK) {
    statusSet(status, __func__, __FILE__, __LINE__, ret);
    return(ret);
  }
  /* Write the pixel data */
  if(verbose>1) {printf("opening file for pixel data\n"); fflush(stdout);}
  FILE *fp;
  if(img->oformat==IMG_FORMAT_NIFTI_1D) {
    fp=fopen(imgfile, "wb");
  } else { // if(img->oformat==IMG_FORMAT_NIFTI_1S
    fp=fopen(imgfile, "r+b");
  }
  if(fp==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_CANNOT_OPEN);
    return(TPCERROR_CANNOT_OPEN);
  }
  /* Move pointer to the start of pixel data */
  if(fseek(fp, (int)dsr.h.vox_offset, SEEK_SET)!=0) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_CANNOT_WRITE);
    fclose(fp); return(TPCERROR_CANNOT_WRITE);
  }
  /* Allocate memory for pixel matrix */
  unsigned int voxNr=img->dimt*img->dimz*img->dimy*img->dimx;
  float *fdata;
  fdata=(float*)calloc(voxNr, sizeof(float));
  if(fdata==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    fclose(fp); return(TPCERROR_OUT_OF_MEMORY);
  }
  /* Write voxel values as floats */
  float *fptr=fdata;
  for(unsigned int fi=0; fi<img->dimt; fi++)
    for(unsigned int zi=0; zi<img->dimz; zi++)
      for(unsigned int yi=0; yi<img->dimy; yi++)
        for(unsigned int xi=0; xi<img->dimx; xi++, fptr++)
          *fptr=img->m[zi][yi][xi][fi];
  if(verbose>1) {printf("writing pixel data\n"); fflush(stdout);}
  fptr=fdata;
  if(fwrite(fptr, sizeof(float), voxNr, fp) != voxNr) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_CANNOT_WRITE);
    free(fdata); fclose(fp); return(TPCERROR_CANNOT_WRITE);
  }

  free(fdata);
  fclose(fp);

  /* Frame times into SIF */
  if(!siffile[0]) { // no frame times to save
    if(verbose>1) printf("  no frame times to save into SIF\n");
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
    return(TPCERROR_OK);
  }
  {
    if(verbose>1) {printf("making SIF\n"); fflush(stdout);}
    TAC sif; tacInit(&sif);
    /* Try to read existing SIF */
    if(fileExist(siffile) && tacRead(&sif, siffile, status)==TPCERROR_OK && sif.sampleNr==img->dimt)
    {
      /* If SIF was found and frameNr matches with image, then update the SIF contents, 
         but keep counts, in case previous SIF comes with actual count info from scanner */
        ret=imgToSIF(img, &sif, 1, 1, 0, verbose-3);
    } else {
      /* otherwise create SIF contents */
      ret=imgToSIF(img, &sif, 1, 1, 2, verbose-3);
    }
    if(ret!=TPCERROR_OK) {
      if(verbose>0) fprintf(stderr, "  Error: cannot create SIF contents.\n");
      tacFree(&sif);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_X);
      return(TPCERROR_INVALID_X);
    }
    if(verbose>1) {printf("writing SIF\n"); fflush(stdout);}
    fp=fopen(siffile, "w");
    if(fp==NULL) {
      tacFree(&sif);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_CANNOT_OPEN);
      return(TPCERROR_CANNOT_OPEN);
    }
    ret=tacWrite(&sif, fp, TAC_FORMAT_SIF, 0, status);
    fclose(fp); tacFree(&sif);
    if(ret!=TPCERROR_OK) {
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_CANNOT_WRITE);
      return(TPCERROR_CANNOT_WRITE);
    }
  }


  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Copy header information in IMG structure into NIfTI-1 header struct.

   @return enum tpcerror (TPCERROR_OK when successful).
 */
int imgSetNifti1Header(
  /** Pointer to IMG structure from which header information is read. */
  IMG *img,
  /** NIfTI database name to be saved in header. */
  const char *dbname,
  /** pointer to NIfTI-1 header struct to be filled. */
  NIFTI_DSR *dsr,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout. */
  int verbose
) {
  if(verbose>0) {printf("%s(..., %s, ...)\n", __func__, dbname); fflush(stdout);}

  if(img==NULL || imgHasData(img)==0 || dsr==NULL) return(TPCERROR_FAIL);

  /* Find pixel value range */
  float pmin, pmax;
  int ret=imgMinMax(img, &pmin, &pmax);
  if(ret!=TPCERROR_OK) return(ret);
  if(verbose>1) {printf("  min=%g\n  max=%g\n", pmin, pmax); fflush(stdout);}

  /* Set NIfTI byte order to current machines byte order */
  dsr->byte_order=endianLittle();
  /* Initiate header struct with zeroes */
  memset(&dsr->h, 0, sizeof(NIFTI_1_HEADER));
  memset(&dsr->e, 0, sizeof(NIFTI_EXTENDER));

  /* Set header */
  dsr->h.sizeof_hdr=NIFTI1_HEADER_SIZE; // 348
  strcpy(dsr->h.data_type, "");
  strncpy(dsr->h.db_name, dbname, 17);
  dsr->h.extents=16384; // not used in NIfTI, but required for Analyze compatibility
  dsr->h.regular='r'; // not used in NIfTI, but required for Analyze compatibility
  dsr->h.dim_info='\0'; // MRI slice ordering
  if(img->oformat==IMG_FORMAT_NIFTI_1S) {
    dsr->h.vox_offset=NIFTI1_HEADER_SIZE+NIFTI1_HEADER_EXTENDER_SIZE; // 352
    strcpy(dsr->h.magic, "n+1");
  } else {
    dsr->h.vox_offset=0;
    strcpy(dsr->h.magic, "ni1");
  }
  strcpy(dsr->h.aux_file, "");

  /* Image dimension */
  for(int i=0; i<8; i++) dsr->h.dim[i]=1;
  dsr->h.dim[0]=4;
  dsr->h.dim[1]=img->dimx;
  dsr->h.dim[2]=img->dimy;
  dsr->h.dim[3]=img->dimz;
  dsr->h.dim[4]=img->dimt;

  /* Intent */
  dsr->h.intent_code=NIFTI_INTENT_NONE;
  dsr->h.intent_p1=0.0;
  dsr->h.intent_p2=0.0;
  dsr->h.intent_p3=0.0;
  /* Try to save pixel units into intent name (not standard NIfTI) */
  strlcpy(dsr->h.intent_name, unitName(img->cunit), 16);

  /* Save data as floats, so that there is no need to scale */
  dsr->h.datatype=NIFTI_DT_FLOAT;
  dsr->h.bitpix=32; // bits per pixel
  dsr->h.scl_slope=1.0; // data as floats, so no need to scale
  dsr->h.scl_inter=0.0; // data as floats, so no need to scale
  dsr->h.glmax=pmax; // unused in NIfTI
  dsr->h.glmin=pmin; // unused in NIfTI
  dsr->h.cal_max=pmax;
  dsr->h.cal_min=0.0;

  /* Pixel size */
  dsr->h.slice_start=0;
  for(int i=0; i<8; i++) dsr->h.pixdim[i]=0.0;
  // https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/qsform.html
  dsr->h.pixdim[0]=1.0; // Set to either 1.0 or -1.0; default should be 1.0.
  dsr->h.pixdim[1]=img->sizex;
  dsr->h.pixdim[2]=img->sizey;
  dsr->h.pixdim[3]=img->sizez;
  dsr->h.slice_end=0;
  dsr->h.slice_code=0;
  dsr->h.xyzt_units=NIFTI_UNITS_MM+NIFTI_UNITS_SEC;
  dsr->h.slice_duration=0.0;
  dsr->h.toffset=0.0;

  /* Coordinate system */
  dsr->h.qform_code=img->xform[0];
  dsr->h.sform_code=img->xform[1];
  dsr->h.quatern_b=img->quatern[0];
  dsr->h.quatern_c=img->quatern[1];
  dsr->h.quatern_d=img->quatern[2];
  dsr->h.qoffset_x=img->quatern[3];
  dsr->h.qoffset_y=img->quatern[4];
  dsr->h.qoffset_z=img->quatern[5];


  /* Description */
  strlcpy(dsr->h.descrip, img->studyNr, 80);
  for(int i=0; i<4; i++) dsr->h.srow_x[i]=img->srow[i];
  for(int i=0; i<4; i++) dsr->h.srow_y[i]=img->srow[4+i];
  for(int i=0; i<4; i++) dsr->h.srow_z[i]=img->srow[8+i];

  /* Extension is left as 0 0 0 0 */

  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
