/** @file niftiio.c
 *  @brief Procedures for NIfTI-1 images.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include "tpcnifti.h"
/*****************************************************************************/

/*****************************************************************************/
/** Verify if specified file name is a NIfTI file.
   @remark Currently only NIfTI-1 is supported.
   @todo Support for NIfTI-2.
   @sa niftiReadHeader, anaExists, micropetExists
   @return Returns 0 if it is not NIfTI, otherwise the version of NIfTI.
 */
int niftiExists(
  /** File name, either header file, image file, or base name without extensions. */
  const char *filename,
  /** If file name refers to a NIfTI file, then header file name will be
      written in this char pointer (space needs to allocated by caller);
      If header and image are combined, then this will be the name of combined file;
      enter NULL if not needed. */
  char *hdrfile,
  /** If file name refers to a NIfTI file, then image file name will be
      written in this char pointer (space needs to allocated by caller);
      If header and image are combined, then this will be the name of combined file;
      enter NULL if not needed. */
  char *imgfile,
  /** If file name refers to a NIfTI file, and if SIF exists, then SIF file name
      will be written in this char pointer (space needs to allocated by caller);
      exter NULL if not needed. */
  char *siffile,
  /** Pointer to NIfTI header, which is filled in this function; enter NULL, if not needed. */
  NIFTI_DSR *header,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {printf("%s(%s, ...)\n", __func__, filename); fflush(stdout);}
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);

  /* Initiate output */
  if(hdrfile!=NULL) hdrfile[0]=(char)0;
  if(imgfile!=NULL) imgfile[0]=(char)0;
  if(siffile!=NULL) siffile[0]=(char)0;

  /* Empty file name means not a NIfTI */
  if(strnlen(filename, 2)<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_FILENAME);
    return(0);
  }

  /* Construct the base file name wo extensions */
  char basefile[FILENAME_MAX]; strlcpy(basefile, filename, FILENAME_MAX); 
  filenameRmExtensions(basefile); if(verbose>1) printf("  basefile := %s\n", basefile);

  /* Combined header and image file exists? */
  int combined=0;
  char temp[FILENAME_MAX+10], localhdrfile[FILENAME_MAX];
  localhdrfile[0]=(char)0;
  strcpy(temp, basefile); strcat(temp, ".nii");
  if(fileExist(temp)) {
    if(verbose>1) printf("  %s exists.\n", temp);
    /* Preserve header and image file names */
    strcpy(localhdrfile, temp);
    if(hdrfile!=NULL) strlcpy(hdrfile, temp, FILENAME_MAX);
    if(imgfile!=NULL) strlcpy(imgfile, temp, FILENAME_MAX);
    combined=1;
  } else {
    if(verbose>1) printf("  %s does not exist.\n", temp);
    /* Not combined file, therefore check that header file exists */
    strcpy(temp, basefile); strcat(temp, ".hdr");
    if(!fileExist(temp)) {
      strcpy(temp, basefile); strcat(temp, ".img.hdr");
      if(!fileExist(temp)) {
        if(verbose>1) printf("  hdr file not found.\n");
        statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_MISSING_HEADER);
        return(0);
      }
    }
    /* Preserve file name of the header */
    strlcpy(localhdrfile, temp, FILENAME_MAX);
    if(hdrfile!=NULL) strlcpy(hdrfile, temp, FILENAME_MAX);
    if(verbose>1) printf("  %s is found.\n", localhdrfile);
    /* Not combined file, therefore check that image file exists */
    strcpy(temp, basefile); strcat(temp, ".img");
    if(!fileExist(temp)) {
      if(verbose>1) printf("  %s not found.\n", temp);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_FILE);
      return(0);
    }
    /* Preserve image filename */
    if(imgfile!=NULL) strlcpy(imgfile, temp, FILENAME_MAX);
    if(verbose>1) printf("  %s is found.\n", temp);
  }
  if(verbose>1) {printf("  combined := %d\n", combined); fflush(stdout);}

  /* Read header to check that this indeed is NIfTI */
  {
    NIFTI_DSR *dsr, local_dsr; if(header==NULL) dsr=&local_dsr; else dsr=header;
    int ret=niftiReadHeader(localhdrfile, dsr, verbose-2);
    if(ret!=TPCERROR_OK) {
      if(verbose>1) {
        printf("  %s was not identified as Nifti header file.\n", localhdrfile); fflush(stdout);}
      statusSet(status, __func__, __FILE__, __LINE__, ret);
      return(0);
    }
    if(verbose>1) {printf("  %s is identified as Nifti.\n", localhdrfile); fflush(stdout);}
  }

  /* SIF exists? */
  strcpy(temp, basefile); strcat(temp, ".sif");
  if(verbose>3) printf("  checking if %s exists\n", temp);
  if(!fileExist(temp)) {
    strcpy(temp, basefile); strcat(temp, ".img.sif");
    if(verbose>3) printf("  checking if %s exists\n", temp);
    if(!fileExist(temp)) {
      strcpy(temp, basefile); strcat(temp, ".nii.sif");
      if(verbose>3) printf("  checking if %s exists\n", temp);
      if(!fileExist(temp)) {
        if(verbose>0) printf("  SIF not found or accessible.\n");
        statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
        return(1); // but otherwise ok NIfTI
      }
    }
  }
  /* Preserve SIF filename */
  if(siffile!=NULL) strcpy(siffile, temp);
  if(verbose>1) {printf("  %s is found.\n", temp); fflush(stdout);}

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return(1);
}
/*****************************************************************************/

/*****************************************************************************/
/** Read Nifti header contents, including 4-byte extender. 
   @remark Currently does not read Nifti-1 header extension.
   @todo Support for NIfTI-2.
   @sa niftiExists, anaReadHeader, micropetHeaderRead
   @return enum tpcerror (TPCERROR_OK when successful).
 */
int niftiReadHeader(
  /** Name of file to read (including path and extension). */
  const char *filename,
  /** Pointer to previously allocated header structure. */
  NIFTI_DSR *dsr,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout */
  int verbose
) {
  if(verbose>0) {printf("%s(%s, ...)\n", __func__, filename); fflush(stdout);}

  /* Check arguments */
  if(strnlen(filename, 2)<1) return(TPCERROR_INVALID_FILENAME);
  if(dsr==NULL) return(TPCERROR_FAIL);

  /* Is current platform little endian (1) or not (0) ? */
  int little=endianLittle();
  if(verbose>2) {
    if(little) printf("little endian platform\n"); else printf("big endian platform\n");
  }

  /* Open file */
  FILE *fp=fopen(filename, "rb"); if(fp==NULL) return(TPCERROR_CANNOT_OPEN);

  /* Try to read the size of header, which should be stored in the beginning as
     4-byte integer in both NIfTI and Analyze header files. */
  int hdrSize=0;
  {
    char buf[4];
    if(fread(buf, 4, 1, fp)<1) {fclose(fp); return(TPCERROR_CANNOT_READ);}
    memcpy(&hdrSize, buf, 4); 
    if(hdrSize!=NIFTI1_HEADER_SIZE && hdrSize!=NIFTI2_HEADER_SIZE) swawbip(&hdrSize, 4);
    if(verbose>1) {printf("  sizeof_hdr := %d\n", hdrSize); fflush(stdout);}
    if(hdrSize==NIFTI2_HEADER_SIZE) {
      if(verbose>1) {printf("  NIfTI-2 header size.\n"); fflush(stdout);}
    } else if(hdrSize==NIFTI1_HEADER_SIZE) {
      if(verbose>1) {printf("  NIfTI-1 or Analyze header size.\n"); fflush(stdout);}
    } else {
      if(verbose>1) {printf("  invalid Nifti sizeof_hdr\n"); fflush(stdout);}
      fclose(fp); return(TPCERROR_INVALID_FORMAT);
    }
  }


  /* Read file into NIfTI header structure */
  if(verbose>1) {printf("  reading header as binary data\n"); fflush(stdout);}
  unsigned char buf[NIFTI1_HEADER_SIZE];
  if(fread(buf, NIFTI1_HEADER_SIZE, 1, fp)<1) {fclose(fp); return(TPCERROR_CANNOT_READ);}
  /* Read NIfTI-1 extender */
  int extender=0;
  for(int n=0; n<4; n++) dsr->e.extension[n]=(char)0;
  if(fread(dsr->e.extension, 4, 1, fp)<1) {
    if(verbose>1) printf("  Nifti header extender not found in %s\n", filename);
    extender=0;
  } else {
    if(verbose>2) printf("  Nifti header extender was found in %s\n", filename);
    extender=1;
  }
  /* Close file */
  fclose(fp);

  /* Get the Nifti Magic number */
  if(buf[344]==(char)0 || buf[344]=='0') {
    if(verbose>0) {printf("  probably Analyze header.\n"); fflush(stdout);}
    return(TPCERROR_UNSUPPORTED);
  }
  memcpy(dsr->h.magic, buf+344, 4);
  if(strcasecmp(dsr->h.magic, "ni1")==0) {
    if(verbose>1) {printf("  separate hdr and img files.\n"); fflush(stdout);}
  } else if(strcasecmp(dsr->h.magic, "n+1")==0) {
    if(verbose>1) {printf("  combined hdr and img data.\n"); fflush(stdout);}
  } else {
    if(verbose>1) {printf("  Nifti magic number not found"); fflush(stdout);}
    if(verbose>2) {
      printf("magic := {%d, %d, %d, %d}\n", dsr->h.magic[0], dsr->h.magic[1],
              dsr->h.magic[2], dsr->h.magic[3]);
    } 
    return(TPCERROR_INVALID_FORMAT);
  }
  dsr->h.magic[3]=(char)0; if(verbose>1) printf("  Nifti Magic number := %s\n", dsr->h.magic);
  /* Check that 4-byte header extender was found, if magic number is n+1 */
  if(strcasecmp(dsr->h.magic, "n+1")==0 && extender==0) {
    if(verbose>1) printf("  Nifti n+1 header extender not found\n");
    return(TPCERROR_INVALID_FORMAT);
  }

  /* Determine from dim[0] if file is big or little endian */
  short int s;
  memcpy(&s, buf+40, 2); if(verbose>10) printf("  s := %d\n", s);
  int same_order=0;
  if(s>0 && s<8) { // same order in file and current machine
    dsr->byte_order=little;
    same_order=1;
  } else {
    swabip(&s, 2); if(verbose>10) printf("  s := %d\n", s);
    if(s>0 && s<8) { // opposite order in file and in current machine
      if(little==1) dsr->byte_order=0; else dsr->byte_order=1;
      same_order=0;
    } else {
      if(verbose>1) {printf("  invalid Nifti byte order\n"); fflush(stdout);}
      return(TPCERROR_INCOMPATIBLE_DATA);
    }
  }
  if(verbose>1) {printf("  Nifti byte order := %d\n", dsr->byte_order); fflush(stdout);}

  /* Fill the header structure fields */
  dsr->h.sizeof_hdr=hdrSize;
  memcpy(&dsr->h.data_type, buf+4, 10);
  memcpy(&dsr->h.db_name, buf+14, 18);
  if(!same_order) swawbip(buf+32, 4);
  memcpy(&dsr->h.extents, buf+32, 4);
  if(!same_order) swabip(buf+36, 2);
  memcpy(&dsr->h.session_error, buf+36, 2);
  memcpy(&dsr->h.regular, buf+38, 1);
  memcpy(&dsr->h.dim_info, buf+39, 1);

  /* dim */
  if(!same_order) swabip(buf+40, 16);
  memcpy(dsr->h.dim, buf+40, 16);
  /* intent parameters */
  if(!same_order) swawbip(buf+56, 4);
  memcpy(&dsr->h.intent_p1, buf+56, 4);
  if(!same_order) swawbip(buf+60, 4);
  memcpy(&dsr->h.intent_p2, buf+60, 4);
  if(!same_order) swawbip(buf+64, 4);
  memcpy(&dsr->h.intent_p3, buf+64, 4);
  if(!same_order) swabip(buf+68, 2);
  memcpy(&dsr->h.intent_code, buf+68, 2);

  /*  */
  if(!same_order) swabip(buf+70, 2);
  memcpy(&dsr->h.datatype, buf+70, 2);
  if(!same_order) swabip(buf+72, 2);
  memcpy(&dsr->h.bitpix, buf+72, 2);
  if(!same_order) swabip(buf+74, 2);
  memcpy(&dsr->h.slice_start, buf+74, 2);
  if(!same_order) swawbip(buf+76, 32);
  memcpy(dsr->h.pixdim, buf+76, 32);
  if(!same_order) swawbip(buf+108, 4);
  memcpy(&dsr->h.vox_offset, buf+108, 4);
  if(!same_order) swawbip(buf+112, 4);
  memcpy(&dsr->h.scl_slope, buf+112, 4);
  if(!same_order) swawbip(buf+116, 4);
  memcpy(&dsr->h.scl_inter, buf+116, 4);
  if(!same_order) swabip(buf+120, 2);
  memcpy(&dsr->h.slice_end, buf+120, 2);
  memcpy(&dsr->h.slice_code, buf+122, 1);
  memcpy(&dsr->h.xyzt_units, buf+123, 1);
  if(!same_order) swawbip(buf+124, 4);
  memcpy(&dsr->h.cal_max, buf+124, 4);
  if(!same_order) swawbip(buf+128, 4);
  memcpy(&dsr->h.cal_min, buf+128, 4);
  if(!same_order) swawbip(buf+132, 4);
  memcpy(&dsr->h.slice_duration,buf+132,4);
  if(!same_order) swawbip(buf+136, 4);
  memcpy(&dsr->h.toffset, buf+136, 4);
  if(!same_order) swawbip(buf+140, 4);
  memcpy(&dsr->h.glmax, buf+140, 4);
  if(!same_order) swawbip(buf+144, 4);
  memcpy(&dsr->h.glmin, buf+144, 4);

  /* study description */
  memcpy(&dsr->h.descrip, buf+148, 80);
  /* Auxiliary filename */
  memcpy(&dsr->h.aux_file, buf+228, 24);

  /* Transformation parameters */
  if(!same_order) swabip(buf+252, 2);
  memcpy(&dsr->h.qform_code, buf+252, 2);
  if(!same_order) swabip(buf+254, 2);
  memcpy(&dsr->h.sform_code, buf+254, 2);
  if(!same_order) swawbip(buf+256, 4);
  memcpy(&dsr->h.quatern_b, buf+256, 4);
  if(!same_order) swawbip(buf+260, 4);
  memcpy(&dsr->h.quatern_c, buf+260, 4);
  if(!same_order) swawbip(buf+264, 4);
  memcpy(&dsr->h.quatern_d, buf+264, 4);
  if(!same_order) swawbip(buf+268, 4);
  memcpy(&dsr->h.qoffset_x, buf+268, 4);
  if(!same_order) swawbip(buf+272, 4);
  memcpy(&dsr->h.qoffset_y, buf+272, 4);
  if(!same_order) swawbip(buf+276, 4);
  memcpy(&dsr->h.qoffset_z, buf+276, 4);
  if(!same_order) swawbip(buf+280, 16);
  memcpy(dsr->h.srow_x, buf+280, 16);
  if(!same_order) swawbip(buf+296, 16);
  memcpy(dsr->h.srow_y, buf+296, 16);
  if(!same_order) swawbip(buf+312, 16);
  memcpy(dsr->h.srow_z, buf+312, 16);

  memcpy(&dsr->h.intent_name, buf+328, 16);

  if(verbose>1) {printf("  complete Nifti header was read.\n"); fflush(stdout);}

  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write NIfTI-1 header contents, including 4-byte extender. 
   @remark Currently does not write header extension.
   @todo Support for NIfTI-2.
   @sa niftiExists, nifitReadHeader
   @return enum tpcerror (TPCERROR_OK when successful).
 */
int niftiWriteHeader(
  /** Name of file to read (including path and extension). */
  const char *filename,
  /** Pointer to header structure to write. */
  NIFTI_DSR *dsr,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout */
  int verbose
) {
  if(verbose>0) {printf("%s(%s, ...)\n", __func__, filename); fflush(stdout);}

  /* Check arguments */
  if(strnlen(filename, 2)<1) return(TPCERROR_INVALID_FILENAME);
  if(dsr==NULL) return(TPCERROR_FAIL);

  /* Check magic number */
  if(strncmp(dsr->h.magic, "ni1", 3)!=0 && strncmp(dsr->h.magic, "n+1", 3)!=0)
    return(TPCERROR_INVALID_FORMAT);

  /* Is current platform little endian (1) or not (0) ? */
  int little=endianLittle();
  if(verbose>2) {
    if(little) printf("little endian platform\n"); else printf("big endian platform\n");
  }
  int same_order=0;
  if(little==dsr->byte_order) same_order=1;


  /* Set up buffer with contents of all zeroes */
  unsigned char buf[NIFTI1_HEADER_SIZE];
  memset(buf, 0, sizeof(NIFTI1_HEADER_SIZE));

  /* Copy header contents into the buffer */
  if(verbose>2) printf("  setting write buffer\n");
  unsigned char *bptr=buf+0; memcpy(bptr, &dsr->h.sizeof_hdr, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+4; memcpy(bptr, &dsr->h.data_type, 10);
  bptr=buf+14; memcpy(bptr, &dsr->h.db_name, 18);
  bptr=buf+32; memcpy(bptr, &dsr->h.extents, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+36; memcpy(bptr, &dsr->h.session_error, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+38; memcpy(bptr, &dsr->h.regular, 1);
  bptr=buf+39; memcpy(bptr, &dsr->h.dim_info, 1);

  bptr=buf+40; memcpy(bptr, dsr->h.dim, 16); if(!same_order) swabip(bptr, 16);
  bptr=buf+56; memcpy(bptr, &dsr->h.intent_p1, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+60; memcpy(bptr, &dsr->h.intent_p2, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+64; memcpy(bptr, &dsr->h.intent_p3, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+68; memcpy(bptr, &dsr->h.intent_code, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+70; memcpy(bptr, &dsr->h.datatype, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+72; memcpy(bptr, &dsr->h.bitpix, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+74; memcpy(bptr, &dsr->h.slice_start, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+76; memcpy(bptr, dsr->h.pixdim, 32); if(!same_order) swawbip(bptr, 32);
  bptr=buf+108; memcpy(bptr, &dsr->h.vox_offset, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+112; memcpy(bptr, &dsr->h.scl_slope, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+116; memcpy(bptr, &dsr->h.scl_inter, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+120; memcpy(bptr, &dsr->h.slice_end, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+122; memcpy(bptr, &dsr->h.slice_code, 1);
  bptr=buf+123; memcpy(bptr, &dsr->h.xyzt_units, 1);
  bptr=buf+124; memcpy(bptr, &dsr->h.cal_max, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+128; memcpy(bptr, &dsr->h.cal_min, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+132; memcpy(bptr, &dsr->h.slice_duration, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+136; memcpy(bptr, &dsr->h.toffset, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+140; memcpy(bptr, &dsr->h.glmax, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+144; memcpy(bptr, &dsr->h.glmin, 4); if(!same_order) swawbip(bptr, 4);

  bptr=buf+148; memcpy(bptr, dsr->h.descrip, 80);
  bptr=buf+228; memcpy(bptr, dsr->h.aux_file, 24);
  bptr=buf+252; memcpy(bptr, &dsr->h.qform_code, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+254; memcpy(bptr, &dsr->h.sform_code, 2); if(!same_order) swabip(bptr, 2);
  bptr=buf+256; memcpy(bptr, &dsr->h.quatern_b, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+260; memcpy(bptr, &dsr->h.quatern_c, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+264; memcpy(bptr, &dsr->h.quatern_d, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+268; memcpy(bptr, &dsr->h.qoffset_x, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+272; memcpy(bptr, &dsr->h.qoffset_y, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+276; memcpy(bptr, &dsr->h.qoffset_z, 4); if(!same_order) swawbip(bptr, 4);
  bptr=buf+280; memcpy(bptr, dsr->h.srow_x, 16); if(!same_order) swawbip(bptr, 16);
  bptr=buf+296; memcpy(bptr, dsr->h.srow_y, 16); if(!same_order) swawbip(bptr, 16);
  bptr=buf+312; memcpy(bptr, dsr->h.srow_z, 16); if(!same_order) swawbip(bptr, 16);
  bptr=buf+328; memcpy(bptr, dsr->h.intent_name, 16);
  bptr=buf+344; memcpy(bptr, dsr->h.magic, 4);

  /* Open file for write */
  FILE *fp;
  if(verbose>2) printf("  opening file for write in binary mode\n");
  if(strncmp(dsr->h.magic, "n+1", 3)==0 && fileExist(filename)) {
    /* Existing single file format NIfTI; just edit the header, do not touch the pixel data */
    fp=fopen(filename, "r+b");
  } else {
    /* Otherwise just open for write */
    fp=fopen(filename, "wb");
  }
  if(fp==NULL) return(TPCERROR_CANNOT_OPEN);

  /* Write buffer into the file */
  if(verbose>2) printf("  writing NIfTI header\n");
  if(fwrite(buf, 1, NIFTI1_HEADER_SIZE, fp) != NIFTI1_HEADER_SIZE) {
    fclose(fp); return(TPCERROR_CANNOT_WRITE);
  }

  /* Write extender */
  if(verbose>2) printf("  writing NIfTI extender\n");
  if(fwrite(dsr->e.extension, 1, NIFTI1_HEADER_EXTENDER_SIZE, fp) != NIFTI1_HEADER_EXTENDER_SIZE) {
    fclose(fp); return(TPCERROR_CANNOT_WRITE);
  }

  /* Close the file */
  fclose(fp);

  if(verbose>1) printf("  NIfTI header written.\n");
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
